/*
 * vifctl/vif_config.c
 * $Rev: 1433 $
 *
 * Copyright (C) 2009-2010, Reto Gantenbein <reto.gantenbein@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * @file vif_config.c
 * @brief Implements the virtual interface configuration storage
 *
 * It defines vif_list which is used to store the device configurations. Further
 * it provides functions to create, delete and query virtual interface
 * configurations.
 */

/* INCLUDES *******************************************************************/

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>            // shmget(), shmat(), shmctl(), ...
#include <errno.h>              // EEXIST
#include <stdio.h>              // perror(), printf(), fprintf(), ...
#include <string.h>             // strcmp(), memset(), perror(), ...
#include "vif.h"


/* LOCAL TYPES ****************************************************************/

/** Virtual interface configuration storage layout */
typedef struct vif_list
{
    int8_t vif_total_num;               /**< Number of virtual interfaces */
    vif_info vif_entries[MAX_NUM_VIF];  /**< Virtual interface configuration */
}
vif_list;

#define VIF_LIST_SIZE           sizeof(vif_list)
#define MATCH_ID(id)            ((id) == list_id)


/* LOCAL VARIABLES ************************************************************/

/** Virtual interface list ID */
int list_id = -1;

/** Vif list shared memory pointer */
vif_list *list = NULL;


/* INTERFACE FUNCTIONS ********************************************************/

int vif_config_create(void);
int vif_config_load(void);
int vif_config_list(int, char[MAX_NUM_VIF][IFNAMSIZ + 1]);
vif_info *vif_config_vif_info(int, const char *);
int vif_config_add(int, const char *);
int vif_config_delete(int, const char *);
int vif_config_destory(int);
void vif_config_print(int);

extern int (*vif_debug)(const char *format, ...);
extern int (*vif_error)(const char *format, ...);


/* PRIVATE FUNCTIONS **********************************************************/

/**
 * Initialize shared memory storage for vif_list
 * @param flags Shared memory flags (see: man shmget)
 * @return List ID
 */
int
vif_config_init(int flags)
{
    key_t shm_key = 42;

    // get shared memory id
    if ((list_id = shmget(shm_key, VIF_LIST_SIZE, flags)) == -1)
    {
        vif_debug("vif_config_init(): shmget (%s)\n", strerror(errno));
        return -1;
    }

    // attach to address space
    if ((list = (vif_list *) shmat(list_id, NULL, 0)) == (vif_list *) -1)
    {
        vif_debug("vif_config_init(): shmat(%i) (%s)\n",
            list_id, strerror(errno));
        return -1;
    }

    vif_debug("vif_config_init(): shmget (list_id = %i)\n", list_id);
    return list_id;
}

/**
 * Return index of given virtual interface
 */
int
vif_config_device_index(const char *dev_name)
{
    int i = 0;
    int index = -1;

    if (list == NULL)
    {
        vif_error("vif_config_device_index(): No configuration loaded\n");
        return -1;
    }
    for (i = 0; i < list->vif_total_num; i++)
        if (strcmp(list->vif_entries[i].vif_name, dev_name) == 0)
            index = i;

    return index;
}


/* VIF INITIALIZATION FUNCTIONS ********(**************************************/

/**
 * Define driver parameter range in allusion to the ath_pci driver. Still a lot
 * of members are uninitialized but they generally are of minor importance.
 *
 * TODO: Function could be extended with driver type as argument, so we can
 *       load the range according to wlan driver
 */
void
vif_config_init_iwrange(iwrange *wlan_range)
{
    int i = 0;

    wlan_range->min_nwid = 0;
    wlan_range->max_nwid = 0;
    wlan_range->sensitivity = 3;                    // ?
    wlan_range->min_frag = 256;
    wlan_range->max_frag = 2346;
    wlan_range->we_version_compiled = WIRELESS_EXT;
    wlan_range->we_version_source = VIF_WE_VERSION; // for wireless-tools compatibility
    wlan_range->num_channels = VIF_NUM_CHANNEL;
    wlan_range->num_frequency = VIF_NUM_FREQ;
    for (i = 0; i < VIF_NUM_FREQ; i++)
    {
        iw_float2freq(channel_frequency[i] * MEGA, &(wlan_range->freq[i]));
        wlan_range->freq[i].i = i;
    }
    wlan_range->min_retry = 0;
    wlan_range->max_retry = VIF_MAX_RETRY;
    wlan_range->retry_capa = IW_RETRY_LIMIT;
    wlan_range->retry_flags = IW_RETRY_LIMIT;
    wlan_range->min_rts = 0;
    wlan_range->max_rts = VIF_MAX_RTS;
}

/**
 * Set default wireless configuration (see iw_compat.h for struct wireless_info)
 * Parameters are choosen as they were observed from real Atheros module
 * loading before any configuration step. Still some values may not correspond
 * to the kernel module initialization.
 * @see vif.h
 */
void
vif_config_init_wlan(const char *dev_name)
{
    vif_info *vif_entry = vif_config_vif_info(list_id, dev_name);
    struct wireless_info *wlan_info = &vif_entry->vif_wlaninfo;
    if (wlan_info == NULL)
        return;

    // set wireless name
    strcpy(wlan_info->b.name, "IEEE 802.11b");

    // initialize driver defaults
    vif_config_init_iwrange(&(wlan_info->range));

    // set nickname (default: none)
    wlan_info->has_nickname = 0;
    wlan_info->nickname[0] = '\0';

    // set AP address (default: none)
    wlan_info->has_ap_addr = 0;
    memset((char *) wlan_info->ap_addr.sa_data, 0x00, ETH_ALEN);

    // power management settings (default: off)
    wlan_info->power.value = 0;
    wlan_info->power.disabled = 1;
    wlan_info->power.fixed = 1;
    wlan_info->has_power = iw_get_param_state(&(wlan_info->power));

    // set sensitivity to custom value 1
    // (will be translated to -85 mW in WlanModel)
    wlan_info->sens.value = 1;
    wlan_info->sens.disabled = 0;
    wlan_info->sens.fixed = 1;
    wlan_info->has_sens = iw_get_param_state(&(wlan_info->sens));

    // set initial bitrate (default: 0 Mbit/s)
    wlan_info->bitrate.value = 0;
    wlan_info->bitrate.disabled = 1;
    wlan_info->bitrate.fixed = 0;
    wlan_info->has_bitrate = iw_get_param_state(&(wlan_info->bitrate));

    // rts threshold in bytes (default: off)
    wlan_info->rts.value = 0;
    wlan_info->rts.disabled = 1;
    wlan_info->rts.fixed = 1;
    wlan_info->has_rts = iw_get_param_state(&(wlan_info->rts));

    // set fragmentation threshold in bytes (default: off)
    wlan_info->frag.value = 0;
    wlan_info->frag.disabled = 1;
    wlan_info->frag.fixed = 1;
    wlan_info->has_frag = iw_get_param_state(&(wlan_info->frag));

    // set initial tx power to 17dBm
    wlan_info->txpower.value = VIF_DEFAULT_TXPWR;
    wlan_info->txpower.disabled = 0;
    wlan_info->txpower.flags = IW_TXPOW_DBM;
    wlan_info->txpower.fixed = 1;
    wlan_info->has_txpower = iw_get_param_state(&(wlan_info->txpower));

    // set retry limit (default: off)
    wlan_info->retry.value = 0;
    wlan_info->retry.disabled = 1;
    wlan_info->retry.fixed = 1;
    wlan_info->has_retry = iw_get_param_state(&(wlan_info->retry));

    // set network id (default: off)
    wlan_info->b.nwid.value = 0;
    wlan_info->b.nwid.disabled = 1;
    wlan_info->b.nwid.fixed = 1;
    wlan_info->b.has_nwid = iw_get_param_state(&(wlan_info->b.nwid));

    // set initial frequency (default: channel 0)
    wlan_info->b.has_freq = 0;
    wlan_info->b.freq = channel_frequency[0] * MEGA;
    wlan_info->b.freq_flags = IW_FREQ_AUTO;

    // initial key settings (default: encoding disabled)
    wlan_info->b.has_key = 0;
    wlan_info->b.key_size = 0;
    wlan_info->b.key[0] = '\0';
    wlan_info->b.key_flags = IW_ENCODE_DISABLED;

    // set initial mode (default: infra)
    wlan_info->b.has_mode = 1;
    wlan_info->b.mode = IW_MODE_INFRA;

    // essid settings (default: none)
    wlan_info->b.has_essid = 0;
    wlan_info->b.essid_on = 1;
    wlan_info->b.essid[0] = '\0';
}


/* PUBLIC FUNCTIONS ***********************************************************/

/**
 * Create a new empty virtual interface list
 * @return List ID (-1 on error)
 */
int
vif_config_create(void)
{
    int i = 0;

    int shm_flags = 0644 | IPC_CREAT;

    if (vif_config_init(shm_flags) < 0)
        return -1;

    list->vif_total_num = 0;
    for (i = 0; i < MAX_NUM_VIF; i++)
        memset(&list->vif_entries[i], 0, sizeof(vif_info));

    //>>>>>>>>>> DEBUG <<<<<<<<<<<<<<
    //~ vif_debug("vif_config_create(): list_id = %i, size = %li, address = %p\n",
        //~ list_id, VIF_LIST_SIZE, list);

    return list_id;
}

/**
 * Load the shared memory virtual interface list
 * @return List ID (-1 on error)
 */
int
vif_config_load(void)
{
    int shm_flags = 0644;

    vif_config_init(shm_flags);

    //>>>>>>>>>> DEBUG <<<<<<<<<<<<<<
    //~ printf("vif_config_load(): list_id = %i, size = %li, address = %p\n",
        //~ list_id, VIF_LIST_SIZE, list);

    return list_id;
}

/** Unload the virtual interface list */
int
vif_config_unload(int id)
{
    if (!MATCH_ID(id))
        return -1;

    if (shmdt(list) < 0)
    {
        vif_error("vif_config_unload(%i): shmdt (%s)", id, strerror(errno));
        return -1;
    }

    return 0;
}

/**
 * Delete the shared memory vif list.
 * @param id List ID
 */
int
vif_config_destory(int id)
{
    if (!MATCH_ID(id))
        return -1;

    return shmctl(id, IPC_RMID, NULL);
}

/**
 * Delete a specific vif configuration from the list
 * @param id List ID
 * @param dev_name Interface device name
 */
int
vif_config_delete(int id, const char *dev_name)
{
    int i = 0, j = 0;
    int vif_index = -1;
    vif_info *vif_entry = NULL;

    if (!MATCH_ID(id))
        return -1;

    for (i = 0; i < list->vif_total_num; i++)
    {
        vif_entry = &list->vif_entries[i];
        if (strncmp(dev_name, vif_entry->vif_name, IFNAMSIZ) == 0)
        {
            vif_index = i;
            memset(vif_entry, 0, sizeof(vif_info));     // overwrite with zero
            if (i < list->vif_total_num - 1)
            {
                for (j = i; j < list->vif_total_num; j++)   // cleanup
                    memmove(&list->vif_entries[j], &list->vif_entries[j + 1],
                            sizeof(vif_info));
            }
            break;
        }
    }
    if (vif_index < 0)
    {
        vif_debug("vif_config_delete(): interface %s not found\n", dev_name);
        return -1;
    }
    if (list->vif_total_num-- == 0)         // no interfaces left
        vif_config_destory(id);             // destroy shared memory area

    return 0;
}

/**
 * List the configured virtual interfaces
 * @param id List ID
 * @param vif_names Char-array to store the virtual interface names
 * @return Number of configured virtual interfaces
 */
int
vif_config_list(int id, char vif_names[MAX_NUM_VIF][IFNAMSIZ + 1])
{
    int i = 0;
    vif_info *vif_entry = NULL;

    if (!MATCH_ID(id))
        return -1;

    if (vif_names != NULL)
    {
        for (i = 0; i < list->vif_total_num; i++)
        {
            vif_entry = &list->vif_entries[i];
            strcpy(vif_names[i], vif_entry->vif_name);  // we already have proper strings
        }
    }

    return list->vif_total_num;
}

/**
 * Return a virtual interface configuration from the list
 * @param id List ID
 * @param dev_name Interface device name
 * @return Pointer to vif_info structure of given device
 */
vif_info *vif_config_vif_info(int id, const char *dev_name)
{
    int vif_index = -1;

    if (!MATCH_ID(id))
        return NULL;

    vif_index = vif_config_device_index(dev_name);
    if (vif_index < 0)
        return NULL;
    else
        return &list->vif_entries[vif_index];
}

/**
 * Add a new default virtual interface configuration to the list
 * @param id List ID
 * @param dev_name Interface device name
 * @return 0 if success, -1 else
 * @see vif_config_init_wlan
 */
int vif_config_add(int id, const char *dev_name)
{
    vif_info *vif_entry = NULL;

    if (!MATCH_ID(id))
        return -1;

    if (vif_config_device_index(dev_name) >= 0)
    {
        vif_debug("vif_config_add(): device %s already exists\n");
        return -1;
    }

    // fill vif information
    vif_entry = &list->vif_entries[list->vif_total_num];
    list->vif_total_num++;
    strncpy(vif_entry->vif_name, dev_name, IFNAMSIZ);
    vif_entry->vif_name[IFNAMSIZ] = '\0';   // properly terminate string
    vif_config_init_wlan(dev_name);

    return 0;
}

/**
 * Print vif_list to the debug output
 * @param id List ID
 */
void
vif_config_print(int id)
{
    int i = 0;
    vif_info *vif_entry = NULL;

    if (!MATCH_ID(id))
    {
        vif_error("Couldn't find valid vif configuration\n");
        return;
    }
    vif_debug("Virtual interface configuration list:\n");
    vif_debug("| vif_total_num = %i\n", list->vif_total_num);
    for (i = 0; i < list->vif_total_num; i++)
    {
        vif_entry = &(list->vif_entries[i]);
        vif_debug("| vif_entries[%i]->vif_name = %s\t\t(address: %p)\n",
            i, (char *) &(vif_entry->vif_name), &(vif_entry->vif_name));
        vif_debug("| vif_entries[%i]->vif_mac = %2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x\t(address: %p)\n",
            i, vif_entry->vif_mac[0], vif_entry->vif_mac[1],
            vif_entry->vif_mac[2], vif_entry->vif_mac[3],
            vif_entry->vif_mac[4], vif_entry->vif_mac[5],
            &(vif_entry->vif_mac));
        //~ vif_debug("|   vif_wlaninfo->txpower.value = %i\t\t(address: %p)\n", vif_entry->vif_wlaninfo.txpower.value, &vif_entry->vif_wlaninfo.txpower.value);
        //~ vif_debug("|   vif_wlaninfo->txpower.fixed = %u\n", vif_entry->vif_wlaninfo.txpower.fixed);
        //~ vif_debug("|   vif_wlaninfo->txpower.disabled = %u\n", vif_entry->vif_wlaninfo.txpower.disabled);
        //~ vif_debug("|   vif_wlaninfo->txpower.flags = %u\n", vif_entry->vif_wlaninfo.txpower.flags);
    }
}
