/*
 * wlanmodel/EmulationRTScheduler.cc
 *
 * Copyright (C) 2008-2010 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

// This file is based on the cSocketRTScheduler.cc of OMNeT++ written by
// Andras Varga.

#include <netdb.h>
#include "EmulationRTScheduler.h"

Register_Class(EmulationRTScheduler);
Register_GlobalConfigOption(CFGID_EMULATIONRTSCHEDULER_PORT, "emulation-port", CFG_INT, "2424", "When EmulationRTScheduler is selected as scheduler class: the port number the emulation listens for traffic.");
Register_GlobalConfigOption(CFGID_EMULATIONRTSCHEDULER_SCALING, "realtime-scaling", CFG_DOUBLE, NULL, "When EmulationRTScheduler is selected as scheduler class: ratio of simulation time to real time.");

/**
 * Constructor
 */
EmulationRTScheduler::EmulationRTScheduler() : cScheduler()
{
    listenerSocket = INVALID_SOCKET;
    connSocket = INVALID_SOCKET;
}

/**
 * Destructor
 */
EmulationRTScheduler::~EmulationRTScheduler()
{
}

/**
 * Called at the beginning of a simulation run.
 */
void
EmulationRTScheduler::startRun()
{
    if (initsocketlibonce()!=0)
        throw new cRuntimeError("EmulationRTScheduler: Cannot initialize socket library");

    factor = ev.getConfig()->getAsDouble(CFGID_EMULATIONRTSCHEDULER_SCALING);
    if (factor!=0)
        factor = 1/factor;
    doScaling = (factor!=0);

    gettimeofday(&baseTime, NULL);

    module = NULL;
    notificationMsg = NULL;
    recvBuffer = NULL;
    recvBufferSize = 0;
    numBytesPtr = NULL;

    port = ev.getConfig()->getAsInt(CFGID_EMULATIONRTSCHEDULER_PORT);
    setupListener();
}

/**
 * Called at the end of a simulation run.
 */
void
EmulationRTScheduler::endRun()
{
    close(listenerSocket);
    close(connSocket);
    listenerSocket = INVALID_SOCKET;
    connSocket = INVALID_SOCKET;
}

/**
 * Setup listening socket for receiving node traffic
 */
void
EmulationRTScheduler::setupListener()
{
	struct addrinfo request;
	struct addrinfo *result;

	memset(&request, 0, sizeof(struct addrinfo));
	request.ai_family = AF_UNSPEC;		// allow IPv4 and IPv6
	request.ai_socktype = SOCK_DGRAM;	// UDP socket
	request.ai_flags = AI_PASSIVE;		// for wildcard IP address
	request.ai_protocol = 0;			// any protocol
	request.ai_canonname = NULL;
	request.ai_addr = NULL;
	request.ai_next = NULL;

	char cport[5];						// max 65536
	sprintf(cport, "%i", port);

    if (getaddrinfo(NULL, cport, &request, &result) < 0)
        throw new cRuntimeError("EmulationRTScheduler: getaddrinfo");

    listenerSocket = socket(result->ai_family, result->ai_socktype, result->ai_protocol);
    if (listenerSocket == INVALID_SOCKET)
        throw new cRuntimeError("EmulationRTScheduler: cannot create socket");

    if (bind(listenerSocket, result->ai_addr, result->ai_addrlen) == SOCKET_ERROR)
        throw new cRuntimeError("EmulationRTScheduler: socket bind() failed");
}

/**
 * Called every time the user hits the Run button in Tkenv.
 * Use this callback to pin current simulation time to current wall clock time.
 */
void
EmulationRTScheduler::executionResumed()
{
    gettimeofday(&baseTime, NULL);
    baseTime = timeval_substract(baseTime, SIMTIME_DBL(doScaling ? factor*sim->getSimTime() : sim->getSimTime()));
}

void
EmulationRTScheduler::setInterfaceModule(cModule *mod, cMessage *notifMsg, \
    char *buf, int bufSize, int *nBytesPtr)
{
    if (module)
        throw new cRuntimeError("EmulationRTScheduler: setInterfaceModule() already called");
    if (!mod || !notifMsg || !buf || !bufSize || !nBytesPtr)
        throw new cRuntimeError("EmulationRTScheduler: setInterfaceModule(): arguments must be non-NULL");

    module = mod;
    notificationMsg = notifMsg;
    recvBuffer = buf;
    recvBufferSize = bufSize;
    numBytesPtr = nBytesPtr;
    *numBytesPtr = 0;
}

/**
 * Listen for incoming network traffic until timeout exceeds
 */
bool EmulationRTScheduler::receiveWithTimeout(long usec)
{
    // prepare sets for select()
    fd_set readFDs, writeFDs, exceptFDs;
    FD_ZERO(&readFDs);
    FD_ZERO(&writeFDs);
    FD_ZERO(&exceptFDs);

    FD_SET(listenerSocket, &readFDs);

    timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = usec;

    if (select(FD_SETSIZE, &readFDs, &writeFDs, &exceptFDs, &timeout) > 0)
    {
        if (listenerSocket != INVALID_SOCKET && FD_ISSET(listenerSocket, &readFDs))
        {
            // receive from listenerSocket
            sockaddr_in sinRemote;
            int addrSize = sizeof(sinRemote);
            char *bufPtr = recvBuffer + (*numBytesPtr);
            int bufLeft = recvBufferSize - (*numBytesPtr);
            if (bufLeft <= 0)
                throw new cRuntimeError("EmulationRTScheduler: interface module's recvBuffer is full");
            int nBytes = recvfrom(listenerSocket, bufPtr, bufLeft, 0, \
                (sockaddr*) &sinRemote, (socklen_t*) &addrSize);
            if (nBytes == SOCKET_ERROR)
            {
                EV << "EmulationRTScheduler: error receiving socket" << sock_errno() << "\n";
                connSocket = INVALID_SOCKET;
            } else {
                // schedule notificationMsg for the interface module
                EV << "EmulationRTScheduler: " << nBytes << " bytes received\n";
                (*numBytesPtr) += nBytes;

                notificationMsg->setName("rtmsg");

                timeval curTime;
                gettimeofday(&curTime, NULL);
                curTime = timeval_substract(curTime, baseTime);
                simtime_t t = curTime.tv_sec + curTime.tv_usec * 1e-6;
                // TBD assert that it's somehow not smaller than previous event's time
                notificationMsg->setArrival(module,-1,t);
                notificationMsg->setTimestamp(t);

                simulation.msgQueue.insert(notificationMsg);
                return true;
            }
        }
    }
    return false;
}

int
EmulationRTScheduler::receiveUntil(const timeval& targetTime)
{
    // if there's more than 200ms to wait, wait in 100ms chunks
    // in order to keep UI responsiveness by invoking ev.idle()
    timeval curTime;
    gettimeofday(&curTime, NULL);
    while (targetTime.tv_sec-curTime.tv_sec >=2 ||
           timeval_diff_usec(targetTime, curTime) >= 200000)
    {
        if (receiveWithTimeout(100000)) // 100ms
            return 1;
        if (ev.idle())
            return -1;
        gettimeofday(&curTime, NULL);
    }

    // difference is now at most 100ms, do it at once
    long usec = timeval_diff_usec(targetTime, curTime);
    if (usec>0)
        if (receiveWithTimeout(usec))
            return 1;
    return 0;
}

/**
 * The scheduler function -- it should return the next event to be processed.
 * Normally (with sequential execution) it just returns msgQueue.peekFirst().
 * With parallel and/or real-time simulation, it is also the scheduler's task
 * to synchronize with real time and/or with other partitions.
 */
cMessage *EmulationRTScheduler::getNextEvent()
{
    timeval targetTime, curTime, diffTime;
    EV << "Processing next event\n";

    // calculate target time
    cMessage *msg = sim->msgQueue.peekFirst();
    if (!msg)
    {
        // if there are no events, wait until something comes from outside
        // TBD: obey simtimelimit, cpu-time-limit
        targetTime.tv_sec = LONG_MAX;
        targetTime.tv_usec = 0;
    }
    else
    {
        // use time of next event
        simtime_t eventSimtime = msg->getArrivalTime();
        targetTime = timeval_add(baseTime, eventSimtime.dbl());
    }

    // if needed, wait until that time arrives
    gettimeofday(&curTime, NULL);
    if (timeval_greater(targetTime, curTime))
    {
        int status = receiveUntil(targetTime);
        if (status == -1)
            return NULL; // interrupted by user
        if (status == 1)
            msg = sim->msgQueue.peekFirst(); // received something
    }
    else
    {
        // we're behind -- customized versions of this class may
        // alert if we're too much behind, whatever that means
        diffTime = timeval_substract(curTime, targetTime);
        EV << "EmulationRTScheduler: We are " << diffTime.tv_sec + diffTime.tv_usec * 1e-6 << " seconds behind\n";
    }

    // ok, return the message
    return msg;
}

int EmulationRTScheduler::sendMessage(const char *msg,
								      const size_t numBytes,
								      const struct sockaddr *nodeAddr,
								      const socklen_t addrSize)
{
    ssize_t bytesSent = sendto(listenerSocket, msg, numBytes, 0, nodeAddr, addrSize);
    EV << "EmulationRTScheduler: " << bytesSent << " bytes sent ";

    if (bytesSent == (ssize_t)numBytes)
    {
        EV << "\n";
        return 0;
    } else {
        EV << "(" << strerror(errno) << ")\n";
        return SOCKET_ERROR;
    }
}
